/*
 * @(#)washra_form.cs        1.0 12/26/2007
 * 
 * Copyright (c) 2007 Senamirmir Project. All Rights Reserved.
 *
 * This file is part of Software: WashRa 4.0
 *
 * WashRa 4.0 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * WashRa is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using washra;

namespace washra_transformer {
   /**
    * This class provides the GUI for the application. Effort was made that its task is not
    * mixed with the business logic--namely the washralib library which does the transformation. 
    * The class is responsible among other things, prompting the user for the document to be
    * converted and passing the file name as an argument to the transformer by instantiating 
    * an object from the WashRaTransformer class. This is done in the "invokeTransformer()"
    * method.
    * 
    * Note: you will notice through out the code that is commented in C++ and C comment format.
    *       Since IDE generated code are commented using C++ comment, it helps to differentiate
    *       that by using C comment for the rest.
    * 
    */

   public partial class WashRaForm : Form {
      private BackgroundWorker backgroundWorker;

      /** constructs ui and setup backgroundworker thread */
      public WashRaForm() {
         InitializeComponent();

         backgroundWorker = new BackgroundWorker();
         backgroundWorker.WorkerReportsProgress = true;
         backgroundWorker.DoWork += new DoWorkEventHandler(backgroundWorker1_DoWork);
         backgroundWorker.RunWorkerCompleted += new RunWorkerCompletedEventHandler(backgroundWorker1_RunWorkerCompleted);
         backgroundWorker.ProgressChanged += new ProgressChangedEventHandler(backgroundWorker1_ProgressChanged);
      }

      private void backgroundWorker1_DoWork(object sender, DoWorkEventArgs e) {
         // This method will run on a thread other than the UI thread.
         // Be sure not to manipulate any Windows Forms controls created
         // on the UI thread from this method.
         backgroundWorker.ReportProgress(0, "");
      
         /** obtains document name */
         String inFile = getDocument();

         /** if no document is obtained, terminate this task */
         if (inFile == null)
            return;

         /** updates progressbar */
         backgroundWorker.ReportProgress(50, "Working...");

         /** invokes the methods that coordinates and performs the task */
         invokeTransformer(inFile);

         /** updates progressbar */         
         backgroundWorker.ReportProgress(100, "Complete!");

         /** displays message with the converted document name */
         postMessage(inFile);
      }

      /** invoked by the BackgroundWorker framework when ReportProgress(...) is called */
      private void backgroundWorker1_ProgressChanged(object sender, ProgressChangedEventArgs e) {
         conversionStripProgressBar.Value = e.ProgressPercentage;
         toolStripStatusLabel.Text        = e.UserState as String;
      }

      /** invoked by the BackgroundWorker framework when ReportProgress(...) is called */
      private void backgroundWorker1_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e) {
         conversionStripProgressBar.Enabled = false;
         convert.Enabled = true;
         exit.Enabled = true;
         convert.Focus();
      }

      /** prompts user to choose document and obtains name */
      String getDocument() {
         OpenFileDialog dlg = new OpenFileDialog();
         dlg.FileName = "Document";
         dlg.DefaultExt = ".rtf";
         dlg.Filter = "Rich Text Format (.rtf) | *.rtf";
         string inFile = null;
         DialogResult result = dlg.ShowDialog(this);

         /** if file is selected, then obtain the file name */
         if (result == DialogResult.OK) {
            inFile = dlg.FileName;
         } else {
            return null;
         }
         /** if no file is selected or wrong file is selected, terminate */
         if (inFile == null) {
            return null;
         } else if ( !inFile.EndsWith(".rtf") ) {
            MessageBox.Show("Document must be in RTF format", "File Error", 
               MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1, MessageBoxOptions.DefaultDesktopOnly);
            return null;
         }
         return inFile;
      }

      /** launches the process to conver the document */
      private void invokeTransformer(String inFile) {
         /** the WashRaTransformer object performs the conversion */
         try {
            new WashRaTransformer(inFile);
         } catch (IOException ioe) {
            MessageBox.Show("Conversion Failure", "Document Error",
               MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1,MessageBoxOptions.DefaultDesktopOnly);
         }
      }

      /** Notifies the output document file name */
      void postMessage(String inFile) {
         /** the output file is figured and displayed */
         inFile = inFile.Insert(inFile.ToLower().IndexOf(".rtf"), "_output");
         MessageBox.Show("Output file is: " + inFile, "Conversion Successful!",
            MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1,MessageBoxOptions.DefaultDesktopOnly);
      }

      /** handles the "Convert" button events */
      private void convert_Click(object sender, EventArgs e) {
         convert.Enabled = false;
         exit.Enabled = false;

         /** sets progressBar attributes and makes it visible*/
         conversionStripProgressBar.Enabled = true;
         //conversionStripProgressBar.Visible = true;
         conversionStripProgressBar.Minimum = 0;
         conversionStripProgressBar.Maximum = 100;
         /** invoke another thread to do the actual work */
         backgroundWorker.RunWorkerAsync();
      }

      /** handles the "Exit" button events */
      private void exit_Click(object sender, EventArgs e) {
         Application.Exit();
      }

      private void WashRaWindow_Load(object sender, EventArgs e) {}
      private void statusBar_ItemClicked(object sender, ToolStripItemClickedEventArgs e) {}
   }
}