/*
 * @(#)WashRaTransformer.cs        1.0 12/26/2007
 * 
 * Copyright (c) 2007 Senamirmir Project. All Rights Reserved.
 *
 * This file is part of Software: WashRa 4.0
 *
 * WashRa 4.0 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * WashRa is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */


using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;
namespace washra {
   /**
    * Microsoft Word documents (others) written using WashRa 1.0, 2.0, 3.0, 3.1, and 3.2 are not 
    * compatible to Unicode, at least, those text formatted with WashRa Ethiopic fonts. The purpose 
    * of this class is to convert such documents that exist in RTF format to a format that is 
    * compatible to Unicode-Ethiopic standard. 
    * 
    * Any document to be processed, it must be in RTF format. There are several versions of The 
    * RTF specification, but Unicode support was introduced starting from version 1.6; therefore,
    * documents that don't confirm to RTF specification 1.6 or above cannot be processed.
    * 
    * It worth mentioning here that application or tools capable of saving contents in RTF format 
    * may produce character representation that is slightly different from another; even within MS 
    * Office application, you do observe this. For instance, if you generate an RTF document using 
    * MS Word 2003, the ASCII code 149 will be saved as '\bullet' while the same character is saved 
    * as '\'95' in Open Office Writer. Very difficult what to expect from a given RTF if it is 
    * generated by various program. This problem is manifested with ASCII code between 146--151 and 
    * maybe more.
    * 
    * Generally, this is how the conversion problem is handled.
    * 
    * 1. A given RTF document has two parts--header and body. The header consists of font, style, 
    *    layout, other definitions. The font definitions are extracted and those related to WashRa 
    *    fonts are converted to WashRa fonts compatible to Unicode. At the same time, the character 
    *    set are changed to ANSI or \fcharset0 type.
    * 
    * 2. The body of the RTF text which consists of the text content is parsed and converted as the 
    *    following:
    * 
    *    a. If the text is formatted with non-WashRa's fonts, conversion doesn't take place.
    *    b. If the text is formatted with WashRa's fonts, each character is evaluated and converted 
    *       to Ethiopic Unicode character. This is done with the following assumption.
    * 
    *       i.   WashRa fonts from version 1.0, 2.0, 3.0, and 3.1 are a symbol type and text formatted 
    *            with a symbol font is saved in special hexdecimal format; for instance, the character 
    *            'A' (ASCII code 65) is saved as \'41. 
    * 
    *       ii.  If a text is formated with non-symbol fonts; that is, ANSI or similar, the lower block 
    *            ASCII characters are saved as is while the upper block of ASCII characters are saved 
    *            in special hexadecimal format using the format \'xx, where xx is the code of the 
    *            character in hexadecimal.
    * 
    *       iii. Some of the characters between ASCII code 146 and 151, maybe saved in control format 
    *            with character name, such as, \bullet or \lquote and so on.
    *    
    */
   public class WashRaTransformer {
      // used to keep track of non-WashRa font used within the RTF document
      List<string> fontDefList = new List<string>();
      // partial multi-dimensional font conversion table which will be completed during conversion
      string[,] cft = new string[,] { 
         {"", "Washra  Primary;", "Ethiopic WashRa SemiBold;"},
         {"", "Washrax Secondary;", "Ethiopic WashRa SemiBold;"},

         {"", "Washrasl  Primary Slanted;", "Ethiopic WashRa SemiBold Slant;"},
         {"", "Washraxsl Secondary Slanted;", "Ethiopic WashRa SemiBold Slant;"},

         {"", "Ethiopia Primary;", "Ethiopic WashRa SemiBold;"},
         {"", "Ethiopia Secondary;", "Ethiopic WashRa SemiBold;"},

         {"", "Ethiopia Primary Slanted;", "Ethiopic WashRa SemiBold Slant;"},
         {"", "Ethiopia Secondary Slanted;", "Ethiopic WashRa SemiBold Slant;"},

         {"", "Wookianos Primary;", "Ethiopic Wookianos;"},
         {"", "Wookianos Secondary;", "Ethiopic Wookianos;"},

         {"", "YebSe Primary;", "Ethiopic YebSe;"},
         {"", "YebSe Secondary;", "Ethiopic YebSe;"},

         {"", "AEthiopia Jiret Set I;", "Ethiopia Jiret;"},
         {"", "AEthiopia Jiret Set II;", "Ethiopia Jiret;"},

         {"", "AEthiopia Set I;", "Ethiopic WashRa SemiBold;"},
         {"", "AEthiopia Set II;", "Ethiopic WashRa SemiBold;"},

         {"", "AHiwua Set I;", "Ethiopic Hiwua;"},
         {"", "AHiwua Set II;", "Ethiopic Hiwua;"},

         {"", "ATint  Set I;", "Ethiopic Tint;"},
         {"", "ATint Set II;", "Ethiopic Tint;"},

         {"", "AWookianos Set I;", "Ethiopic Wookianos;"},
         {"", "AWookianos Set II;", "Ethiopic Wookianos;"},

         {"", "AYebSe Set I;", "Ethiopic YebSe;"},
         {"", "AYebSe Set II;", "Ethiopic YebSe;"}
      };


      /**
       * Reads the input document and stashes it into StringBuilder object. 
       * 
       * Parameters
       *    inFile: the input RTF document to be converted 
       */
      public WashRaTransformer(String inFile) {
         try {
            /** if this file doesn't have an RTF extension, return */
            int index = inFile.IndexOf(".rtf");
            if (index < 0) {
               throw new IOException();
            }

            /** creates file stream object */
            FileStream fsr = new FileStream(inFile, FileMode.Open, FileAccess.Read);
            /** creates reader stream object */
            StreamReader sr = new StreamReader(fsr);
            /** reads the file content into a StringBuilder instance */
            StringBuilder rtfText = new StringBuilder(sr.ReadToEnd());
            /** closes the infile */
            sr.Close();

            /** constructs font tables for conversion purpose */
            processFontTable(rtfText);
            /** performs the conversion */
            convertEthiopicText(rtfText);

            /** create a new file name from the infile by inserting _output before extension */
            String outFile = inFile.Insert(inFile.IndexOf(".rtf"), "_output");
            /** creates a file stream object */
            FileStream fsw = new FileStream(outFile, FileMode.Create, FileAccess.Write);
            /** creates a writer stream object */
            StreamWriter sw = new StreamWriter(fsw);
            /** writes the converted rtfText into an output file */
            sw.Write(rtfText.ToString());
            /** close the outfile */
            sw.Close();
            
            /** these write stmts are for console environment */
            Console.WriteLine();
            Console.WriteLine("Done!");
         } catch (IOException ie) {
            Console.WriteLine(ie);
         }      
      }

      /**
       * Converts either char or character controls into Ethiopic-Unicode char. When a file
       * is saved in RTF format under MS Word, the characters are encoded as the following.
       * 
       * 1. If the characters are symbols (the font type is symbol), the characters are
       *    encoded using this escape sequence: \'xx (xx--denotes the character code in
       *    Hexadecimal).
       * 2. If the characters are from ANSI compatible fonts, the lower character set from
       *    0-127 are coded as is while from 128-255 are codes as:  \'xx.
       * 
       * The conversion is done using WPConversionTable, WSConversionTable, ESIConversionTable,
       * and ESIIConversionTable classes which consist of two static dimensional arrays that
       * map from ASCII--Unicode.
       * 
       * Parameters
       *    rtfText: the the content of the RTF text
       *    token:   the character or character control to be converted
       *    codeMap: conversion table
       *    index:   the index where the character or character-control is located in rtfText
       */
      int convertCharControl(StringBuilder rtfText, string token, string[,] codeMap, int index) {
         /** Note: for some reason these ANSI codes produce an extra space when a document is converted 
          *  into RTF format. I don't know why, but after the document is transformed into Unicode that 
          *  extra space becomes hard white space breaking a word. This problem is corrected by removing that
          *  extra space during transformation as the following.
          */
         if (token.Equals(@"\endash") || token.Equals(@"\emdash") || token.Equals(@"\rquote") || token.Equals(@"\bullet"))
            rtfText.Replace(" ", @"", index + token.Length, 1);

         // obtain the matching index for conversion
         int row = getCodeMapIndex(codeMap, token);
         // if matching index for conversion is found, do the conversion
         if (row >= 0) {        
            rtfText.Replace(codeMap[row, 0], codeMap[row, 1] + @"\'3f", index, codeMap[row, 0].Length);
            index = index + codeMap[row, 1].Length - 1 + @"\'3f".Length; // adjust i to the correct index after the update
         }
         return index;
      }


      /**
       * Returns a single character control from the current position from rtfText. When a document
       * with ASNI font and the characters 146, 147, 149, 150, and 151, are saved in RTF format
       * under MS Word, they are saved as the following character controls
       * 
       * Parameters
       *    rtfText     : the content of the RTF text
       *    startIndex  : the current position where the character control starts
       *    controlType : a reference type to notify the caller if this is character control or other 
       *                  RTF control
       */
      string getControl(StringBuilder rtfText, int startIndex, ref int controlType) {
         int index = startIndex;
         // if this is rtf hexa char control , then it should be four char long with format: \'xx
         if (rtfText[startIndex] == '\\' && rtfText[startIndex + 1] == '\'') {
            controlType = 1;
            //Console.WriteLine("getControl --> " + rtfText.ToString(startIndex, 4));
            return rtfText.ToString(startIndex, 4);
         }

         // this delimiters is used for lookup 
         string delimiters = "  \t\r\n{\\};";
         // used to construct the new control
         string newControl = "";
         // used to track control status
         bool control = false;
         // parses and constructs the control
         for (int i = startIndex; i < rtfText.Length; i++) {
            if (rtfText[i] == '\\') {
               newControl += rtfText[i];
               control = true;
            } else if ((control == true) && (i + 1 < rtfText.Length) && (delimiters.IndexOf(rtfText[i + 1]) >= 0)) {
               control = false;
               newControl += rtfText[i];
               if (newControl.Equals("\\-"))
                  controlType = 1;
               else if (newControl.Equals("\""))
                  controlType = 1;
               else if (newControl.Equals("\\endash"))
                  controlType = 1;
               else if (newControl.Equals("\\emdash"))
                  controlType = 1;
               else if (newControl.Equals("\\lquote"))
                  controlType = 1;
               else if (newControl.Equals("\\rquote"))
                  controlType = 1;
               else if (newControl.Equals("\\bullet"))
                  controlType = 1;
               else
                  controlType = 0;
               break;
            } else {
               newControl += rtfText[i];
            }
         }
         return newControl;
      }


      /**
       * Sets a proper conversion table based on the beggining of the rtf text format. This
       * methods must be called after the processFontTable() func is called.
       */
      string[,] setDefaultCodeConversionTable() {
         int i = 0;
         // check the font table which consists of all font definition from the RTF \fonttbl
         for (; i < cft.Length / 3; i++) {
            if (!cft[i, 0].Equals("")) break;
         }
         // if there is font def below the half table point, it is Washra 3.1. and lower version fonts def
         if (i < (cft.Length / 3) / 2) 
            return WPConversionTable.table;
         // if there is font def above the half table point, it is Washra 3.2 fonts def
         else if (i < cft.Length / 3) 
            return ESIConversionTable.table;
         // otherwise not
         else 
            return null;
      }


      /**
       * Used to advance index within rtfText to by a group at a time; a group in RTF is defined
       * by braces ({}).
       */
      int getNewGroup(StringBuilder rtfText, string oldGroup) {
         // matches and determines the index of the control here named as oldGroup
         int index = BMStringMatch.search(rtfText, oldGroup, 0, rtfText.Length);
         // if no control is found, return bad
         if (index < 0) return -1;

         // adjusts the index to open brace by decrementing it and push it to stack
         WStack<int> stack = new WStack<int>();
         stack.push(index - 1);
         // locates the end of the group
         for (; index < rtfText.Length; index++) {
            if (rtfText[index] == '{')
               stack.push(index);
            else if (rtfText[index] == '}')
               stack.pop();
            if (stack.getIndex() == -1)
               return ++index;
         }
         return -1;
      }


      /**
       * Returns the index of a given character or character-control if it is found in code map table.
       * Parameters:
       *    table    : a code mapping table (from WashRa to Ethiopic Unicode)
       *    token    : a character or character-control to be searched
       */
      int getCodeMapIndex(string[,] table, string token) {
         int i = 0;
         for (; i < table.Length / 2; i++) {
            if (table[i, 0].Equals(token)) {
               return i;
            }
         }
         return -1;
      }


      /**
       * Extracts font definition controls from the rtfText and constructs two tables--one for
       * Ethiopic fonts and the other for non-Ethiopic fonts. These tables are used by other
       * functions during conversion.
       * Parameters:
       *    rtfText  : content the document in RTF text format--this StringBuilder object
       */
      void processFontTable(StringBuilder rtfText) {
         // stack used to parse RTF groups delimited by braces
         WStack<int> stack = new WStack<int>();
         // returns the matching str index
         int start = BMStringMatch.search(rtfText, @"{\fonttbl", 0, rtfText.Length);
         int end = start;
         stack.push(start);
         // parses the rtfText and performs conversion
         for (int i = start + 1; stack.getIndex() != -1; i++) {
            if (rtfText[i] == '{') {            // if the current char is '{', pushes its index
               stack.push(i);
            } else if (rtfText[i] == '}') {     // if the current char is '}', pops its matching open index ...
               start = stack.pop();
               end = i;
               // if stack is empty, start afresh from the next char 
               if (stack.getIndex() != 0) continue;

               bool found = false;
               int controlType = 0;
               // from the current RTF group, extracts the 1st control, which is assumed font-def
               string fontControl = getControl(rtfText, start + 1, ref controlType);

               // fills the cft for Ethiopic fonts
               for (int k = 0; k < cft.Length / 3; k++) {
                  int startIndex = BMStringMatch.search(rtfText, cft[k, 1], start, end);
                  if (startIndex >= 0) {
                     // assign the found font definition control to array
                     cft[k, 0] = fontControl;

                     // searches for the replacement Ethiopic
                     startIndex = BMStringMatch.search(rtfText, cft[k, 1], start, end);

                     // if the replacement Ethiopic font is found, do the conversion
                     if (startIndex >= 0)
                        rtfText.Replace(cft[k, 1], cft[k, 2], startIndex, cft[k, 1].Length);

                     // searches if the char set is type of symbol
                     startIndex = BMStringMatch.search(rtfText, @"\fnil\fcharset2", start, end);

                     // if the the char set type is symbol, converts it to ANSI set
                     if (startIndex >= 0)
                        rtfText.Replace(@"\fnil\fcharset2", @"\fnil\fcharset0", startIndex, @"\fnil\fcharset2".Length);

                     // update i to reflect the current position after modification
                     i += cft[k, 2].Length - cft[k, 1].Length;

                     // this fond def is for Ethiopic, thus found should be true
                     found = true;
                     break;
                  } // close inner if
               } // close for

               /* if fontControl is not a def for Ethiopic fonts, staches it into an ArrayList */
               if (!found)
                  fontDefList.Add(fontControl);       // collects non-Ethiopic fonts font-def 
            } // else 
         } // end for loop
      }


      /**
       * Returns true if the given font-def (control) is non-Ethiopic; otherwise, false.
       * Parameters:
       *    fontControl  : RTF control (such as \f10, ...)
       */
      bool inFontDefList(string fontControl) {
         IEnumerable list = fontDefList;
         foreach (string item in list) {
            if (item.Equals(fontControl)) {
               return true;
            }
         }
         return false;
      }


      /**
       * Determines what conversion table should be used next based on the given font-def
       * Parameters:
       *    codeMap      : the current conversion table which might be returned to the caller
       *                   if fontControl isn't found in cft.
       *    fontControl  : RTF control (such as \f10, ...)
       */
      string[,] setConversionTable(string[,] codeMap, string fontControl) {
         for (int j = 0; j < cft.Length / 3; j++) {
            if (cft[j, 0].Equals(fontControl)) {
               switch (j) {
                  case 0:
                  case 2:
                  case 4:
                  case 6:
                  case 8:
                  case 10:
                     return WPConversionTable.table;
                  case 1:
                  case 3:
                  case 5:
                  case 7:
                  case 9:
                  case 11:
                     return WSConversionTable.table;
                  case 12:
                  case 14:
                  case 16:
                  case 18:
                  case 20:
                  case 22:
                     return ESIConversionTable.table;
                  case 13:
                  case 15:
                  case 17:
                  case 19:
                  case 21:
                  case 23:
                     return ESIIConversionTable.table;
               }
            }
         }
         return codeMap;
      }


      /**
       * Converts the Ethiopic text found in the rtfTetxt
       * Parameters:
       *    rtfText      : content the document in RTF text format--this StringBuilder object
       */
      void convertEthiopicText(StringBuilder rtfText) {
         // codeMap is used to perform conversion from WashRa code to Ethiopic Unicode
         string[,] codeMap;
         codeMap = setDefaultCodeConversionTable();
         // advances the index within the rtfText as much close to the body of RTF document
         int start = getNewGroup(rtfText, @"\xmlnstbl");
         start = (start < 0 ? getNewGroup(rtfText, @"\fonttbl") : start);

         string token = "";
         int controlType = 0;
         // processes each WashRa char code or char-control and performs the conversion
         for (int i = (start < 0 ? 0 : start); i < rtfText.Length; i++) {
            // prints something on console while doing the conversion
            if (i % 10000 == 0) Console.Write(".");   

            // Some text groups miss font definition; in that case, set conversion to WashRa Primary table
            if (rtfText[i] == '}') {
               codeMap = setDefaultCodeConversionTable();
            } else if (rtfText[i] == '\\') {
               token = getControl(rtfText, i, ref controlType);

               // if proper font def, get conversion table
               codeMap = setConversionTable(codeMap, token);

               // if the current control is font def and it is the other type, set codeMap = false
               if (controlType == 0 && inFontDefList(token))
                  codeMap = null;

               // if everything is alright, do conversion
               if (controlType == 1 && codeMap != null) {
                  i = convertCharControl(rtfText, token, codeMap, i);
               } else {
                  i += (token.Length - 1);  // adjust i to correct index
               }
               continue;
            } else {
               // determines and adjusts the code table based on font definition
               if (codeMap != null) {
                  token = "" + rtfText[i];
                  i = convertCharControl(rtfText, token, codeMap, i);
               }
            }
         }
      }


      /**
       * An entry point for program execution
       */
      static void Main(string[] args) {
//         new WashRaTransformer(@"C:\old_laptop\senafidel\washra-4.1\scratch\docamara.rtf");
      }
   }
}